@@ -15,6 +15,8 @@
 import uuid
 import random
 import time
+import json
+import concurrent.futures
 from enum import Enum
 from fog05_sdk.yaks_connector import Yaks_Connector
 from fog05_sdk.interfaces import Constants
@@ -23,6 +25,101 @@ from fog05_sdk.interfaces.InfraFDU import InfraFDU
 from mvar import MVar
 
 
+
+class RunningFDU(object):
+    def __init__(self, connector, executor, instanceid, env,  sysid=Constants.default_system_id,
+                                    tenantid=Constants.default_tenant_id):
+        '''
+        Class: RunningFDU
+
+        This class implements the behaviour for a running FDU
+
+        attributes
+        ----------
+        instanceid : string
+            The associated FDU Instance ID
+
+        '''
+        self.sysid =  sysid
+        self.tenantid = tenantid
+        self.connector = connector
+        self.executor = executor
+        self.instanceid = instanceid
+        self.env = env
+        self.started = False
+        self.exit_code = None
+        self.log = None
+        self.var = MVar()
+        self.err_var = None
+
+    def run_job(self):
+        '''
+        Job that waits the FDU run to end
+        '''
+        self.stated = True
+        res = self.connector.glob.actual.run_fdu_in_node(self.sysid, self.tenantid, self.instanceid, self.env)
+        if res.get('error') is not None:
+                self.err_var = res.get('error')
+                self.var.put(-255)
+        else:
+            self.err_var = None
+            self.var.put(int(res.get('result')))
+
+
+    def run(self):
+        '''
+        Submits the run_job
+        '''
+        if self.started is False:
+            self.exit_code = None
+            self.log =  None
+            self.executor.submit(self.run_job)
+        else:
+            raise ValueError('FDU is still running, wait it to finish before restaring')
+
+    def get_result(self):
+        '''
+        If the FDU is still running blocks until in ends, then returns its exit code and log
+
+        ---
+        returns
+        (int, string)
+        '''
+        if self.exit_code == None and self.log == None:
+            exit_code = self.var.get()
+            if self.err_var is not None:
+                raise ValueError(self.err_var)
+            self.err_var = None
+            log = self.connector.glob.actual.log_fdu_in_node(self.sysid, self.tenantid, self.instanceid)
+
+            if log.get('error') is not None:
+                self.exit_code = int(exit_code)
+                raise ValueError(log.get('error'))
+            self.started = False
+            self.exit_code = int(exit_code)
+            self.log = log.get('result')
+        return (self.exit_code, self.log)
+
+    def get_log(self):
+        '''
+        Returns the log of the FDU if it has ended
+        ---
+        returns
+        string
+        '''
+        return self.log
+
+    def get_code(self):
+        '''
+        Returns the exit code of the FDU if it has ended
+        ---
+        returns
+        int
+        '''
+        return self.exit_code
+
+
+
 class FIMAPI(object):
     '''
     Class: FIMAPI
@@ -55,11 +152,12 @@ class FIMAPI(object):
         self.connector = Yaks_Connector(locator)
         self.sysid = sysid
         self.tenantid = tenantid
+        self.executor = concurrent.futures.ThreadPoolExecutor()
         self.descriptor = self.Descriptor()
         self.node = self.Node(self.connector, self.sysid, self.tenantid)
         self.plugin = self.Plugin(self.connector, self.sysid, self.tenantid)
         self.network = self.Network(self.connector, self.sysid, self.tenantid)
-        self.fdu  = self.FDUAPI(self.connector, self.sysid, self.tenantid)
+        self.fdu  = self.FDUAPI(self.connector, self.sysid, self.tenantid, self.executor)
         self.image = self.Image(self.connector, self.sysid, self.tenantid)
         self.flavor = self.Flavor(self.connector, self.sysid, self.tenantid)
 
@@ -364,14 +462,18 @@ class FIMAPI(object):
 
             returns
             -------
-            bool
+            string
             '''
 
             descriptor.update({'status': 'add'})
             net_id = descriptor.get('uuid')
+            if net_id is None:
+                net_id = uuid.uuid4()
+                descriptor['uuid'] = net_id
 
             self.connector.glob.desired.add_network(
                 self.sysid, self.tenantid, net_id, descriptor)
+            return net_id
 
         def remove_network(self, net_uuid):
             '''
@@ -416,7 +518,7 @@ class FIMAPI(object):
                 return net
             res = self.connector.glob.actual.create_network_in_node(self.sysid, self.tenantid, nodeid, descriptor)
             if res.get('error') is not None:
-                raise ValueError('Got Error {}'.format(res['error']))
+                raise ValueError('Got  Error {} with message {}'.format(res['error'], res['error_msg']))
             return res['result']
 
         def remove_network_from_node(self, netid, nodeid):
@@ -436,7 +538,7 @@ class FIMAPI(object):
             '''
             res = self.connector.glob.actual.remove_network_from_node(self.sysid, self.tenantid, nodeid, netid)
             if res.get('error') is not None:
-                raise ValueError('Got Error {}'.format(res['error']))
+                raise ValueError('Got  Error {} with message {}'.format(res['error'], res['error_msg']))
             return res['result']
 
 
@@ -508,7 +610,7 @@ class FIMAPI(object):
             res = self.connector.glob.actual.add_node_port_to_network(self.sysid, self.tenantid, node, port_info['uuid'], net_uuid)
             if res.get('result') is not None:
                 return cp_uuid
-            raise ValueError('Error connecting: {}'.format(res['error']))
+            raise ValueError('Error connecting: {} with message {}'.format(res['error'], res['error_msg']))
 
         def disconnect_cp(self, cp_uuid):
             '''
@@ -536,7 +638,7 @@ class FIMAPI(object):
             res = self.connector.glob.actual.remove_node_port_from_network(self.sysid, self.tenantid, node, port_info['uuid'])
             if res.get('result') is not None:
                 return cp_uuid
-            raise ValueError('Error connecting: {}'.format(res['error']))
+            raise ValueError('Error connecting: {} with message {}'.format(res['error'], res['error_msg']))
 
         def add_router(self, nodeid, descriptor):
             '''
@@ -738,13 +840,14 @@ class FIMAPI(object):
         '''
 
         def __init__(self, connector=None, sysid=Constants.default_system_id,
-                     tenantid=Constants.default_tenant_id):
+                     tenantid=Constants.default_tenant_id, executor=None):
 
-            if connector is None:
-                raise RuntimeError('Yaks connector cannot be none in API!')
+            if connector is None or executor is None:
+                raise RuntimeError('Yaks connector or executor cannot be none in API!')
             self.connector = connector
             self.sysid = sysid
             self.tenantid = tenantid
+            self.executor = executor
 
         def __wait_node_fdu_state_change(self, instanceid, state):
             '''
@@ -781,6 +884,41 @@ class FIMAPI(object):
                     state, instanceid,fdu_info.get('error_code'), fdu_info.get('error_msg')))
             return fdu_info
 
+        def __wait_specific_node_fdu_state_change(self, nodeid, instanceid, state):
+            '''
+            Waits an FDU instance state to change
+
+            parameters
+            ----------
+            instanceid : string
+                UUID of instance
+            state : string
+                new state
+
+            returns
+            --------
+            dictionary
+
+            '''
+
+            fdu_info = self.connector.glob.actual.get_node_fdu_instance(
+                self.sysid, self.tenantid, nodeid, instanceid)
+            while fdu_info is None:
+                    fdu_info = self.connector.glob.actual.get_node_fdu_instance(
+                self.sysid, self.tenantid, nodeid, instanceid)
+            fdu = InfraFDU(fdu_info)
+            es = fdu.get_status()
+            while es.upper() not in [state, 'ERROR']:
+                fdu_info = self.connector.glob.actual.get_node_fdu_instance(
+                self.sysid, self.tenantid, nodeid, instanceid)
+                fdu = InfraFDU(fdu_info)
+                es = fdu.get_status()
+
+            if es.upper() == 'ERROR':
+                raise ValueError('Unable to change state to {} for FDU Instance: {} Errno: {} Msg: {}'.format(
+                    state, instanceid,fdu_info.get('error_code'), fdu_info.get('error_msg')))
+            return fdu_info
+
         def onboard(self, descriptor):
             '''
             Registers an FDU descriptor in the system catalog
@@ -796,7 +934,7 @@ class FIMAPI(object):
             FDU
             '''
             if not isinstance(descriptor, FDU):
-                raise ValueError('descriptor should be of type FDU')
+                raise ValueError('descriptor should be of type FDU; actual type: {}'.format(type(descriptor)))
             nodes = self.connector.glob.actual.get_all_nodes(self.sysid, self.tenantid)
             if len(nodes) == 0:
                 raise SystemError('No nodes in the system!')
@@ -804,7 +942,7 @@ class FIMAPI(object):
 
             res = self.connector.glob.actual.onboard_fdu_from_node(self.sysid, self.tenantid, n, descriptor.get_uuid(), descriptor.to_json())
             if res.get('result') is None:
-                raise SystemError('Error during onboarding {}'.format(res['error']))
+                raise SystemError('Error during onboarding {} with message {}'.format(res['error'], res['error_msg']))
             return FDU(res['result'])
 
 
@@ -825,11 +963,11 @@ class FIMAPI(object):
             res = self.connector.glob.desired.remove_catalog_fdu_info(
                 self.sysid, self.tenantid, fdu_uuid)
             # if res.get('result') is None:
-            #     raise SystemError('Error during onboarding {}'.format(res['error']))
+            #     raise SystemError('Error during onboarding {} with message {}'.format(res['error'], res['error_msg']))
 
             return fdu_uuid
 
-        def define(self, fduid, node_uuid, wait=True):
+        def define(self, fduid, node_uuid=None, wait=True):
             '''
             Defines the given fdu in the given node
 
@@ -840,31 +978,46 @@ class FIMAPI(object):
             fduid : string
                 UUID of the FDU
             node_uuid : string
-                UUID of the node
+                UUID of the node optional
             wait : bool
                 optional, call will block until FDU is defined
             returns
             -------
             InfraFDU
             '''
-            desc = self.connector.glob.actual.get_catalog_fdu_info(
-                self.sysid, self.tenantid, fduid)
-            if desc is None:
-                raise ValueError('FDU with this UUID not found in the catalog')
 
-            res = self.connector.glob.actual.define_fdu_in_node(self.sysid, self.tenantid, node_uuid, fduid)
-            if res.get('error') is not None:
-                raise ValueError('Got Error {}'.format(res['error']))
-            if wait:
-                self.__wait_node_fdu_state_change(res['result']['uuid'],'DEFINE')
-            return InfraFDU(res['result'])
+
+            if node_uuid is not None:
+
+                desc = self.connector.glob.actual.get_catalog_fdu_info(
+                    self.sysid, self.tenantid, fduid)
+                if desc is None:
+                    raise ValueError('FDU with this UUID not found in the catalog')
+
+                res = self.connector.glob.actual.define_fdu_in_node(self.sysid, self.tenantid, node_uuid, fduid)
+                if res.get('error') is not None:
+                    raise ValueError('Got  Error {} with message {}'.format(res['error'], res['error_msg']))
+                if wait:
+                    self.__wait_node_fdu_state_change(res['result']['uuid'],'DEFINE')
+                return InfraFDU(res['result'])
+            else:
+                nodes = self.connector.glob.actual.get_all_nodes(self.sysid, self.tenantid)
+                if len(nodes) == 0:
+                    raise SystemError('No nodes in the system!')
+                n = random.choice(nodes)
+                res = self.connector.glob.actual.schedule_fdu_from_node(self.sysid, self.tenantid, n, fduid)
+                if res.get('error') is not None:
+                    raise ValueError('Got  Error {} with message {}'.format(res['error'], res['error_msg']))
+                if wait:
+                    self.__wait_node_fdu_state_change(res['result']['uuid'],'DEFINE')
+                return InfraFDU(res['result'])
 
 
         def undefine(self, instanceid):
             '''
             Undefines the given instance
 
-            paremeters
+            parameters
             ----------
             instanceid : string
                 UUID of instance
@@ -893,7 +1046,7 @@ class FIMAPI(object):
             '''
             Configures the given instance
 
-            paremeters
+            parameters
             ----------
             instanceid : string
                 UUID of instance
@@ -926,7 +1079,7 @@ class FIMAPI(object):
             '''
             Cleans the given instance
 
-            paremeters
+            parameters
             ----------
             instanceid : string
                 UUID of instance
@@ -956,14 +1109,16 @@ class FIMAPI(object):
                 self.__wait_node_fdu_state_change(instanceid,  'DEFINE')
             return instanceid
 
-        def start(self, instanceid, wait=True):
+        def start(self, instanceid, env="", wait=True):
             '''
             Starts the given instance
 
-            paremeters
+            parameters
             ----------
             instanceid : string
                 UUID of instance
+            env : string
+                Environment variables to be configured at the starting of the FDU in the format MYVAR=MYVALUE,MYVAR2=MYVALUE2,...
             wait : bool
                 optional, call will block until FDU is started
 
@@ -972,29 +1127,59 @@ class FIMAPI(object):
             string
             '''
 
-            node = self.connector.glob.actual.get_fdu_instance_node(
-                self.sysid, self.tenantid, instanceid)
-            if node is None:
-                raise ValueError('Unable to find node for this instanceid')
-
-            record = self.connector.glob.actual.get_node_fdu_instance(
-                self.sysid, self.tenantid, node, instanceid)
-
-            record = InfraFDU(record)
-
-            record.set_status('RUN')
-            fduid = record.get_fdu_id()
-
-            res = self.connector.glob.desired.add_node_fdu(self.sysid, self.tenantid, node, fduid, instanceid, record.to_json())
             if wait:
-                self.__wait_node_fdu_state_change(instanceid,  'RUN')
+                res = self.connector.glob.actual.start_fdu_in_node(self.sysid, self.tenantid, instanceid, env)
+                if res.get('error') is not None:
+                    raise ValueError(res.get('error'))
+                return res.get('result')
+
+            self.executor.submit(self.connector.glob.actual.start_fdu_in_node,self.sysid, self.tenantid, instanceid, env)
             return instanceid
 
+            # node = self.connector.glob.actual.get_fdu_instance_node(
+            #     self.sysid, self.tenantid, instanceid)
+            # if node is None:
+            #     raise ValueError('Unable to find node for this instanceid')
+
+            # record = self.connector.glob.actual.get_node_fdu_instance(
+            #     self.sysid, self.tenantid, node, instanceid)
+
+            # record = InfraFDU(record)
+
+            # record.set_status('RUN')
+            # fduid = record.get_fdu_id()
+
+            # res = self.connector.glob.desired.add_node_fdu(self.sysid, self.tenantid, node, fduid, instanceid, record.to_json())
+            # if wait:
+            #     self.__wait_node_fdu_state_change(instanceid,  'RUN')
+            # return instanceid
+
+        def run(self, instanceid, env=""):
+            '''
+            Runs and waits the given instance unit it ends
+            returns a RunningFDU object where wait the FDU to end its execution.
+
+            The RunningFDU object can be used to re-run the given FDU
+
+            parameters
+            ----------
+            instanceid : string
+                UUID of instance
+            env : string
+                Environment variables to be configured at the starting of the FDU in the format MYVAR=MYVALUE,MYVAR2=MYVALUE2,...
+            returns
+            -------
+            RunningFDU
+            '''
+            res = RunningFDU(self.connector, self.executor, instanceid, env)
+            res.run()
+            return res
+
         def stop(self, instanceid, wait=True):
             '''
             Stops the given instance
 
-            paremeters
+            parameters
             ----------
             instanceid : string
                 UUID of instance
@@ -1028,7 +1213,7 @@ class FIMAPI(object):
             '''
             Pauses the given instance
 
-            paremeters
+            parameters
             ----------
             instanceid : string
                 UUID of instance
@@ -1062,7 +1247,7 @@ class FIMAPI(object):
             '''
             Resumes the given instance
 
-            paremeters
+            parameters
             ----------
             instanceid : string
                 UUID of instance
@@ -1097,7 +1282,7 @@ class FIMAPI(object):
             '''
             Migrates the given instance
 
-            paremeters
+            parameters
             ----------
             instanceid : string
                 UUID of instance
@@ -1137,7 +1322,7 @@ class FIMAPI(object):
                                                 node, fduid, instanceid, src_record.to_json())
 
             if wait:
-                self.__wait_node_fdu_state_change(instanceid, 'RUN')
+                self.__wait_specific_node_fdu_state_change(destination_node_uuid, instanceid, 'RUN')
             return instanceid
 
 
@@ -1176,7 +1361,7 @@ class FIMAPI(object):
 
             This function calls: stop, clean, undefine
 
-            paremeters
+            parameters
             ----------
             instanceid : string
                 UUID of instance
@@ -1191,6 +1376,59 @@ class FIMAPI(object):
             self.clean(instanceid)
             return self.undefine(instanceid)
 
+        def log(self, instanceid):
+            '''
+            Gets the log for the given FDU instance
+            parameters
+            ----------
+            instanceid : string
+                UUID of instance
+
+            returns
+            -------
+            string
+            '''
+            log = self.connector.glob.actual.log_fdu_in_node(self.sysid, self.tenantid, instanceid)
+
+            if log.get('error') is not None:
+                raise ValueError(log.get('error'))
+            return log.get('result')
+
+        def ls(self, instanceid):
+            '''
+            Lists the file in the given FDU instance directory
+            parameters
+            ----------
+            instanceid : string
+                UUID of instance
+
+            returns
+            -------
+            string list
+            '''
+            ls = self.connector.glob.actual.ls_fdu_in_node(self.sysid, self.tenantid, instanceid)
+
+            if ls.get('error') is not None:
+                raise ValueError(ls.get('error'))
+            return json.loads(ls.get('result'))
+
+        def get_file(self, instanceid, filename):
+            '''
+            Gets the given filename for the given FDU instance
+            parameters
+            ----------
+            instanceid : string
+                UUID of instance
+
+            returns
+            -------
+            string
+            '''
+            data = self.connector.glob.actual.file_fdu_in_node(self.sysid, self.tenantid, instanceid, filename)
+
+            if data.get('error') is not None:
+                raise ValueError(data.get('error'))
+            return data.get('result')
 
         def search(self, search_dict, node_uuid=None):
             '''
@@ -1348,6 +1586,9 @@ class FIMAPI(object):
             '''
 
             img_id = descriptor.get('uuid')
+            if img_id is None:
+                img_id = uuid.uuid4()
+                descriptor['uuid'] = img_id
             res = self.connector.glob.desired.add_image(self.sysid,
              self.tenantid,img_id, descriptor)
             return img_id
@@ -1444,8 +1685,11 @@ class FIMAPI(object):
             -------
             string
             '''
-
             flv_id = descriptor.get('uuid')
+            if flv_id is None:
+                flv_id = uuid.uuid4()
+                descriptor['uuid'] = flv_id
+
             res = self.connector.glob.desired.add_flavor(self.sysid,
              self.tenantid,flv_id, descriptor)
             return flv_id
@@ -1516,3 +1760,23 @@ class FIMAPI(object):
             string list
             '''
             raise NotImplementedError("Not yet...")
+
+
+class FIMAuthExcetpion(Exception):
+    def __init__(self, message):
+        super(FIMAuthExcetpion, self).__init__(message)
+
+
+class FIMAResouceExistingException(Exception):
+    def __init__(self, message):
+        super(FIMAResouceExistingException, self).__init__(message)
+
+
+class FIMNotFoundException(Exception):
+    def __init__(self, message):
+        super(FIMNotFoundException, self).__init__(message)
+
+
+class FIMTaskFailedException(Exception):
+    def __init__(self, message):
+        super(FIMTaskFailedException, self).__init__(message)
\ No newline at end of file
